//******************************************************************************
/// \author     Oliver Fortmeier, SC RWTH-Aachen
//******************************************************************************
/// \brief      Compute the 1-norm (col-sum-norm) and the max-norm
///             (row-sum-norm) of a square matrix.
/**             In order to investigate the cache effects multiple versions
                of this norm-caclulation has been written. The C-sytle is used
                to allocate the matrix, so the matrix is stored so-called
                row-wise.
*/
//******************************************************************************
/// \file       norm.cpp
//******************************************************************************
//  last mod:   (Tim Cramer, RZ RWTH-Aachen) 18.03.2012
//******************************************************************************

/*
    The one norm of a (n x m) matrix is defined as:
        ||A||_1        := max_{j=1..m} sum_{i=1}^n a_ij

    The maximum norm of a matrix is defined as:
        ||A||_infinity := max_{i=1..n} sum_{j=1}^m a_ij
*/

#include <iostream>
#include <iomanip>
#include <cmath>
#include <stdlib.h>
#include <float.h>

#include "realtime.h"        // for timing

namespace norm{

/// \brief Absoulte value of a number
inline double abs(double x)
/// Using this abs function instead of the std::fabs function allows us to
/// control inlining better.
/// \return absolut value of x
{
    return x<0 ? -x : x;
}

/// \brief Initialise a n times n matrix
void init(double** A, const int n)
/** Initialise a matrix by value -1 in the block (0..n-2)x(0..n-2) and
    increasing values in the last row and last column, so the local max variable
    in the norm calculation has to be updated in each step.
\param A The matrix that has to be initialised
\param n dimension of the square matrix */
{
    // init upper left block
    for (int i=0; i<n-1; ++i)
        for (int j=0; j<n-1; ++j)
            A[i][j]=-1.;

    // init last row and column
    for (int i=0; i<n; ++i)
    {
        A[n-1][i]=-(i+2);
        A[i][n-1]=-(i+2);
    }
}

///\brief Allocate row wise memory for the matrix
void allocate_matrix(double*** A, const int n){
/** Allocate a field for each row of the Matrix and a field of pointers
    for each beginning of a row.
\param A The pointer to the matrix that has to be allocated
\param n dimension of the square matrix */
   *A = new double*[n];
    if (!(*A)){
        std::cerr << "Cannot allocate enough memory!" << std::endl;
        exit(1);
    }

    for (int i=n-1; i>=0; i--){
        (*A)[i] = new double[n];
        if (!(*A)[i]){
            std::cerr << "Cannot allocate enough memory!" << std::endl;
            exit(1);
        }
    }
}

///\brief Frees the memory for the row wise the matrix
void destroy_matrix(double** A, const int n){
/** Frees the field which has been allocated for each row of the Matrix and 
    the field of row pointers.
\param A The matrix that has to be deleted 
\param n dimension of the square matrix */
    for (int i=0; i<n; ++i){
        delete[] A[i];
    }
    delete[] A;
}


/// \brief Maximum norm of a matrix
double norm_max(double** const A, const int n)
/** Compute the maximum norm by going through the rows of a matrix. The matrix
    is stored in C-style, so this procedure should work quite well.
\param  A the matrix
\param  n dimension of the square matrix
\return maximum norm of A */
{
    double rowsum=0., max_norm=-1.;
    for (int i=0; i<n; ++i)
    {
        rowsum=0.;
        for (int j=0; j<n; ++j)
        {
            rowsum += abs(A[i][j]); 
        }
        if (rowsum>max_norm)
            max_norm=rowsum;
    }
    return max_norm;
}


/// \brief One Norm of a matrix
double norm1_col_wise(double** const A, const int n)
/** Compute the one norm of a matrix by going through the colums of the matrix.
    The rows of the matrix are lying consecutive in the memory.
\param A the matrix
\param n dimension of the square matrix
\return 1-norm of A*/
{
    double colsum=0., max_norm=-1.;
    for (int j=0; j<n; ++j)
    {
        colsum=0.;
        for (int i=0; i<n; ++i)
        {
            //TODO: Calculate the column sum here. Hint: use the abs() function!
        }
        //TODO: Update the max_norm, if the colsum is bigger
    }
    return max_norm;
}

/// \brief One Norm of a matrix
double norm1_row_wise(double** const A, const int n)
/** Compute the one norm of a matrix by going through the rows of the matrix and
    storing temporary sums in an auxillary array. This should be more efficient
    as the first version, because the rows of the matrix are lying consecutive
    in the memory.
\param A the matrix
\param n dimension of the square matrix
\return 1-norm of A*/
{
    double *colsums= new double[n];
    double max_norm=-1.;

		
    //TODO: Init helper array with 0
    
    //TODO: Iterate the matrix in the correct order and 
    //calculate the colsums using the helper array
    
    //TODO: Search for the max_norm in the array colsums

    delete[] colsums;
    return max_norm;
}

}   // end of namespace norm


using namespace std;
using norm::init;
using norm::allocate_matrix;
using norm::destroy_matrix;
using norm::norm_max;
using norm::norm1_row_wise;
using norm::norm1_col_wise;

int main (int argc, char *argv[] )
{
    int n;      // size of the matrix
    double flops;
    double** A; // the matrix

    if (argc==1){
        cout << "Using default values: matrix dimension 5000" << endl;
        n=5000;
    }
    else if (argc==2){
        n=atoi(argv[1]),      // size of the matrix
        cout << "Using matrix dimension " << n << endl;
    }
    else{
        cerr<<"Usage: "<<argv[0]<<" <matrix_dimension> <blocking_size>"<<endl;
        return 1;
    }

    cout << "Allocate memory " << (n*n*sizeof(double))/1024./1024. << " MB for storing the matrix" << endl;
    allocate_matrix(&A,n);
    // init matrix
    init(A,n);

    // time measurement (with OpenMP-routines)
    double time_start, t1, t2, t3;
    const int num_tests=5;

    // results
    double n1, n2, n3;

    flops = n*n;

    // compute the both norms
    cout << "Compute maximum norm"<<endl;
    time_start= GetRealTime();
    for (int i=0; i<num_tests; ++i){
        n1= norm_max(A, n);
    }
    t1= (GetRealTime() - time_start)/num_tests;

    cout << "Compute 1-norm (column-wise)"<<endl;
    time_start= GetRealTime();
    for (int i=0; i<num_tests; ++i){
        n2= norm1_col_wise(A, n);
    }
    t2= (GetRealTime() - time_start)/num_tests;

    cout << "Compute 1-norm (row-wise)"<<endl;
    time_start= GetRealTime();
    for (int i=0; i<num_tests; ++i){
        n3= norm1_row_wise(A, n);
    }
    t3= (GetRealTime() - time_start)/num_tests;

    destroy_matrix(A, n);

    cout << "The norm are: "
         << "\n - maximum norm   "<<n1<<" in "<<t1<<" sec (" << flops/t1/1e6 << " MFlops)";
    if(abs(n1-n2)<=DBL_EPSILON)
        cout << "\n - 1-norm (col)   "<<n2<<" in "<<t2<<" sec (" << flops/t2/1e6 << " MFlops)";
    else
        cout << "\n - 1-norm (col)   FAILED";
    if(abs(n1-n3)<=DBL_EPSILON)
        cout << "\n - 1-norm (row)   "<<n3<<" in "<<t3<<" sec (" << flops/t3/1e6 << " MFlops)";
    else
        cout << "\n - 1-norm (row)   FAILED";
    cout << endl;

    return 0;
}
