//******************************************************************************
/// \author     Tim Cramer, RZ RWTH-Aachen, cramer@rz.rwth-aachen.de
//******************************************************************************
/// \brief      Executes and measures the performance of various vector-matrix-
///             multiplications (vxm) for different problem sizes.
/*              Each operation is executed several times to get accurate 
                timings. Depending on the error check, the performance is 
                printed out in Mflops.
*/
//******************************************************************************
/// \file       main.c
//******************************************************************************
//  last mod:   (Tim Cramer, RZ RWTH-Aachen) 11.03.2011
//******************************************************************************

#include "lab.h"

double flops;

const int dimensions[] = {100, 250, 500, 1000, 2000, 4000, 5000, 10000};
const int repeat_count = 10;


int main(int argc, char **argv){
	int       m, n;
	long long kr;
	double    memsize;
	double    *y, *A, *x, *ref;
	int       run, ndimensions;
	double    t0, t1, dt;


	printf("%*s %9s%9s%9s%9s\n", 30, 
		"(values in Mflops)",
		"Ref",
		"Ordered",
		"Restrict",
		"MKL"
	);


	ndimensions = sizeof(dimensions) / sizeof(*dimensions);

	for(run=0; run<ndimensions; run++){
		m = n = dimensions[run];

		flops   = 2.0*m*n;

		memsize  = sizeof(double) * (m + m * n + n); //Bytes

		allocate_storage(m, n, &y, &A, &x, &ref);
		init_data(m, n, y, A, x, ref);

		printf("\n(%1d.) %5dx%5d (%7.2f MB): ", run+1, m, n, memsize/1e6);


		// Execute and measure the reference solution
		t0 = GetRealTime();
		for(kr = 0; kr<repeat_count; kr++)
			vxm_ref(m, n, A, x, y);
		t1 = GetRealTime();
		dt = (t1-t0) / repeat_count;

		if(check_results(m, n, y, ref)>0)
			printf("%9s", "FAILED");
		else
			printf("%9.2f", (flops / dt) / 1e6);

		// init data
		init_data(m, n, y, A, x, ref);


		// Execute and measure the orderd solution
		t0 = GetRealTime();
		for(kr=0; kr<repeat_count; kr++)
			vxm_order(m, n, A, x, y);
		t1 = GetRealTime();
		dt = (t1-t0) / repeat_count;

		if(check_results(m, n, y, ref)>0)
			printf("%9s", "FAILED");
		else
			printf("%9.2f", (flops / dt) / 1e6);

		// init data
		init_data(m, n, y, A, x, ref);


		// Execute and measure the restricted solution
		t0 = GetRealTime();
		for (kr=0; kr<repeat_count; kr++)
			vxm_restrict(m, n, A, x, y);
		t1 = GetRealTime();
		dt = (t1-t0) / repeat_count;

		if (check_results(m, n, y, ref) > 0)
			printf("%9s", "FAILED");
		else
			printf("%9.2f", (flops / dt) / 1e6);

		// init data
		init_data(m, n, y, A, x, ref);


		// Execute and measure the BLAS solution
		t0 = GetRealTime();
		for(kr=0; kr<repeat_count; kr++)
			vxm_mkl(m, n, A, x, y);
		t1 = GetRealTime();
		dt = (t1-t0) / repeat_count;

		if (check_results(m, n, y, ref) > 0)
			printf("%9s", "FAILED");
		else
			printf("%9.2f", (flops / dt) / 1e6);

		// Clean up
		free(y);
		free(A);
		free(x);
		free(ref);
	}
	
	printf("\n");

	return 0;
}
